/*-----------------------------------------------------------------------------
 *  "gcc-plugin.h" must be the FIRST file to be included 
 *-----------------------------------------------------------------------------*/
#include "gcc-plugin.h"
#include "config.h"
#include <stdio.h>
#include "system.h"
#include "coretypes.h"
#include "tm.h"
//#include "tm_p.h"
#include "diagnostic.h"
#include "gimple-pretty-print.h"
#include "tree-flow.h"
#include "tree-pass.h"
#include "toplev.h"
#include "gimple.h"

#include "vec.h"
#include "ggc.h"
/*-----------------------------------------------------------------------------
 *  Each plugin MUST define this global int to assert compatibility with GPL; 
 *  else the compiler throws a fatal error 
 *-----------------------------------------------------------------------------*/
int plugin_is_GPL_compatible;

/* global declarations */

/* Count of local entities */
int number_of_nodes = 0;
static int dfs_num = 0;
static int *block_dfs = NULL;

static unsigned int gimple_dfn(void);
static void dfs_numbering_of_bb_inner_recursive (basic_block bb);

/*-----------------------------------------------------------------------------
 *  Structure of the pass we want to insert, identical to a regular ipa pass
 *-----------------------------------------------------------------------------*/

struct gimple_opt_pass pass_plugin = 
{
  {
    GIMPLE_PASS,
    "gimple-dfn",                          /*  name */
    NULL,                                 /*  gate */
    gimple_dfn,                      /*  execute */
    NULL,                                 /*  sub */
    NULL,                                 /*  next */
    0,                                    /*  static pass number */
    TV_INTEGRATION,                        /*  tv_id */
    0,                                    /*  properties required */
    0,                                    /*  properties provided */
    0,                                    /*  properties destroyed */
    0,                                    /*  todo_flags start */
    0                                     /*  todo_flags end */
  }
};

/*-----------------------------------------------------------------------------
 *  This structure provides the information about inserting the pass in the
 *  pass manager. 
 *-----------------------------------------------------------------------------*/
struct register_pass_info pass_info = {
  &(pass_plugin.pass),                  /* Address of new pass, here, the 'struct
                                        opt_pass' field of 'simple_ipa_opt_pass'
                                        defined above */
        "cfg",                          /* Name of the reference pass for hooking up
                                        the new pass.   */
        0,                              /* Insert the pass at the specified instance
                                        number of the reference pass. Do it for
                                        every instance if it is 0. */
        PASS_POS_INSERT_AFTER           /* how to insert the new pass: before,
                                        after, or replace. Here we are inserting
                                        a pass names 'plug' after the pass named
                                        'cfg' */
};

/*-----------------------------------------------------------------------------
 *  plugin_init is the first function to be called after the plugin is loaded
 *-----------------------------------------------------------------------------*/
  int
plugin_init(struct plugin_name_args *plugin_info,
    struct plugin_gcc_version *version)
{

        /*-----------------------------------------------------------------------------
        * Plugins are activiated using this callback 
        *-----------------------------------------------------------------------------*/
        register_callback (
        plugin_info->base_name,     /* char *name: Plugin name, could be any
                                     name. plugin_info->base_name gives this
                                     filename */
        PLUGIN_PASS_MANAGER_SETUP,  /* int event: The event code. Here, setting
                                     up a new pass */
        NULL,                       /* The function that handles event */
        &pass_info);                /* plugin specific data */

        return 0;
}

/* ---------------------------------------------
   The main driver function to perform analysis.
   ---------------------------------------------*/
static unsigned int gimple_dfn(void)
{
  
	number_of_nodes = n_basic_blocks;

  	block_dfs = (int *) ggc_alloc_cleared_atomic (sizeof (int) * number_of_nodes);
	for (int i=0; i < number_of_nodes; ++i)
	{
		block_dfs[i] = -1;
	}
  
	basic_block bb;
	dfs_num = 0;
	bb = ENTRY_BLOCK_PTR;
	dfs_numbering_of_bb_inner_recursive (bb);

	fprintf(dump_file, "DFS : No of nodes : %d\n\n",number_of_nodes);
	fprintf(dump_file, "BB Index\tDFN\n");
	for(int i=0; i<number_of_nodes; ++i)
	{
		fprintf(dump_file,"%d\t\t%d\n",i,block_dfs[i]);
	}
	fprintf(dump_file,"\n\n");
	return 0;
}


static void dfs_numbering_of_bb_inner_recursive (basic_block bb)	/* Recursive version */
{

  edge_iterator ei;
  edge e;
  basic_block succ_bb = NULL;

  block_dfs[bb->index] = dfs_num;
  dfs_num++;

  FOR_EACH_EDGE (e, ei, bb->succs)
  {
	succ_bb = e->dest;
        if (block_dfs[succ_bb->index] == -1 )
	      dfs_numbering_of_bb_inner_recursive (succ_bb);
  }

}
